from figh.base import ConfigStorage
from figh.errors import DupKeyException
from figh.github import ConfigFileFormat, GithubConfigStorage

__all__ = ["ConfigStorage", "GithubConfigStorage", "ConfigFileFormat", "DupKeyException"]



import base64
import os
from typing import Any, Dict, List

import yaml
from github import Auth, Github
from github.ContentFile import ContentFile
from github.Repository import Repository

from figh.base import VERSION_LATEST, ConfigStorage
from figh.config_file_format import ConfigFileFormat
from figh.util import nested_set


class GithubConfigStorage(ConfigStorage):
    github: Github
    repo: Repository
    base_branch: str
    file_format: ConfigFileFormat

    class Config:
        arbitrary_types_allowed = True

    def __init__(
        self,
        repo_name: str,
        token: str,
        base_branch: str = "main",
        file_format: ConfigFileFormat = ConfigFileFormat.YAML,
    ):
        self.repo_name = repo_name
        self.github = Github(auth=Auth.Token(token))
        self.repo = self.github.get_repo(repo_name)
        self.base_branch = base_branch
        self.file_format = file_format

    def _get_repo_contents(self, ref, root_path: str = "/") -> Dict[str, Any]:
        files = self.repo.get_contents(root_path, ref=ref)
        all_files_data = {}
        while files:
            file_content = files.pop(0)
            if file_content.type == "dir":
                files.extend(self.repo.get_contents(file_content.path))
            else:
                content_string = base64.b64decode(file_content.content).decode("utf-8")
                if self.file_format.matched(file_content.path):
                    all_files_data[file_content.path] = content_string
        return all_files_data

    def _merge_configs(
        self, root_path: str, config_files: Dict[str, ContentFile]
    ) -> Dict[str, Any]:
        merged_config = {}
        if self.file_format == ConfigFileFormat.YAML:
            for path, content in config_files.items():
                config_keys = self._file_config_keys(root_path, path)
                config_data = yaml.safe_load(content)
                nested_set(merged_config, config_keys, config_data)
            return merged_config
        else:
            raise ValueError(f"unsupported config file format: {self.file_format}")

    @staticmethod
    def _file_config_keys(root_path: str, file_path: str) -> List[str]:
        root_path = root_path.strip("/")
        file_path = file_path.lstrip("/")
        if file_path.startswith(root_path):
            file_path = file_path[len(root_path) :].lstrip("/")
            file_path = os.path.splitext(file_path)[0]
            items = file_path.split("/")
            items = [item for item in items if item]
            return items
        else:
            return []

    def get(self, version: str = VERSION_LATEST, root_path: str = "/") -> Dict[str, Any]:
        if version is VERSION_LATEST:
            config_files = self._get_repo_contents(ref=self.base_branch, root_path=root_path)
        else:
            tag_ref = self.repo.get_git_ref(f"tags/{version}")
            tag_sha = tag_ref.object.sha
            config_files = self._get_repo_contents(ref=tag_sha, root_path=root_path)
        return self._merge_configs(root_path=root_path, config_files=config_files)



from figh import ConfigFileFormat


def test_config_file_format_matched():
    f = ConfigFileFormat.YAML
    assert f.matched("item.yaml")
    assert f.matched("foo/bar/item.yaml")
    assert f.matched("foo/bar/item.yml")



from abc import ABC, abstractmethod
from typing import Any, Dict

VERSION_LATEST = "latest"


class ConfigStorage(ABC):
    @abstractmethod
    def get(self, version: str = VERSION_LATEST) -> Dict[str, Any]:
        """returns configuration of the version"""



from enum import Enum


class ConfigFileFormat(Enum):
    YAML = "yaml"

    def matched(self, file_path: str) -> bool:
        if self == ConfigFileFormat.YAML:
            return file_path.endswith(".yaml") or file_path.endswith(".yml")
        raise ValueError("unsupported file format to check match")

    def __str__(self):
        return self.value

    def __eq__(self, o):
        return str(self) == str(o)

    def __hash__(self):
        return hash(self.name)



from figh.errors import DupKeyException


def nested_set(dic, keys, value):
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})
    if keys[-1] in dic and dic[keys[-1]]:
        raise DupKeyException(f"key {keys[-1]} duplicated")
    dic[keys[-1]] = value



import os
from typing import Optional

import pytest

from figh import DupKeyException, GithubConfigStorage

TEST_RESOURCES_PATH = "/tests/resources"


@pytest.fixture(scope="module")
def github_test_storage(setup_test_env) -> GithubConfigStorage:
    token = os.getenv("TEST_GITHUB_TOKEN")
    repo = "skspark/figh-py"
    return GithubConfigStorage(repo_name=repo, token=token)


def test_single_file(github_test_storage):
    expected = {"first": {"foo": {"bar": "first"}, "foo2": {"bar2": "second"}}}
    config = github_test_storage.get(root_path=f"{TEST_RESOURCES_PATH}/single-yaml")
    print(f"single-yaml file config: {config}")
    assert config is not None
    assert config == expected


def test_nested(github_test_storage):
    expected = {
        "first": {"foo": {"bar": "first"}},
        "second": {
            "base": {"foo": {"bar": "second"}},
            "fourth": {"base": {"foo": {"bar": "fourth"}}},
        },
        "third": {"base": {"foo": {"bar": "third"}}},
    }
    config = github_test_storage.get(root_path=f"{TEST_RESOURCES_PATH}/nested-yaml")
    print(f"nested-yaml file config: {config}")

    assert config is not None
    assert config == expected


def test_dup_key(github_test_storage):
    dupkey_error: Optional[Exception] = None
    try:
        config = github_test_storage.get(root_path=f"{TEST_RESOURCES_PATH}/duplicated-yaml")
    except DupKeyException as e:
        dupkey_error = e
    assert dupkey_error is not None
